@@ -33,7 +33,7 @@
 void fe_silc_modules_init(void);
 void fe_silc_modules_deinit(void);
 
-void fe_silc_init(void)
+static void silc_init(void)
 {
   theme_register(fecommon_silc_formats);
 
@@ -42,11 +42,21 @@ void fe_silc_init(void)
   fe_silc_messages_init();
   fe_silc_queries_init();
   fe_silcnet_init();
+}
 
+void fe_common_silc_init(void)
+{
+  silc_init();
+  module_register("silc", "fe_common");
+}
+
+void fe_silc_init(void)
+{
+  silc_init();
   module_register("silc", "fe");
 }
 
-void fe_silc_deinit(void)
+void fe_common_silc_deinit(void)
 {
   fe_silc_queries_deinit();
   fe_silc_messages_deinit();
@@ -56,3 +66,8 @@ void fe_silc_deinit(void)
 
   theme_unregister();
 }
+
+void fe_silc_deinit(void)
+{
+  fe_common_silc_deinit();
+}
